import { all } from 'deepmerge';
import 'array.prototype.flatmap/auto.js';
import escapeRegExp from 'escape-string-regexp';
import { AST_NODE_TYPES } from '@typescript-eslint/typescript-estree';
import { ESLintUtils } from '@typescript-eslint/experimental-utils';
import { AST_NODE_TYPES as AST_NODE_TYPES$1 } from '@typescript-eslint/typescript-estree/dist/ts-estree/ast-node-types';

const config = {
    rules: {
        "ts-immutable/immutable-data": "error",
        "ts-immutable/no-class": "error",
        "ts-immutable/no-conditional-statement": "error",
        "ts-immutable/no-expression-statement": "error",
        "ts-immutable/no-let": "error",
        "ts-immutable/no-loop-statement": "error",
        "ts-immutable/no-reject": "error",
        "ts-immutable/no-this": "error",
        "ts-immutable/no-throw": "error",
        "ts-immutable/no-try": "error"
    },
    overrides: [
        {
            files: ["*.ts", "*.tsx"],
            rules: {
                "ts-immutable/no-method-signature": "error",
                "ts-immutable/no-mixed-interface": "error",
                "ts-immutable/prefer-readonly-types": "error"
            }
        }
    ]
};

const config$1 = {
    rules: {
        "prefer-const": "error",
        "no-param-reassign": "error",
        "no-var": "error"
    },
    overrides: [
        {
            files: ["*.ts", "*.tsx"],
            rules: {
                "@typescript-eslint/explicit-function-return-type": [
                    "error",
                    {
                        allowExpressions: true,
                        allowTypedFunctionExpressions: true,
                        allowHigherOrderFunctions: true
                    }
                ]
            }
        }
    ]
};

const config$2 = all([
    config$1,
    {
        rules: {
            "ts-immutable/no-let": "error",
            "ts-immutable/immutable-data": "error"
        },
        overrides: [
            {
                files: ["*.ts", "*.tsx"],
                rules: {
                    "ts-immutable/no-method-signature": "warn",
                    "ts-immutable/prefer-readonly-types": "error"
                }
            }
        ]
    }
]);

const config$3 = all([
    config$2,
    {
        rules: {
            "ts-immutable/no-this": "error",
            "ts-immutable/no-class": "error",
            "ts-immutable/no-loop-statement": "error",
            "ts-immutable/no-conditional-statement": [
                "error",
                { allowReturningBranches: true }
            ],
            "ts-immutable/no-throw": "error",
            "ts-immutable/no-return-void": "error",
            "ts-immutable/functional-parameters": "error"
        },
        overrides: [
            {
                files: ["*.ts", "*.tsx"],
                rules: {
                    "ts-immutable/no-mixed-interface": "error"
                }
            }
        ]
    }
]);

const config$4 = all([
    config$3,
    {
        rules: {
            "ts-immutable/no-conditional-statement": "error",
            "ts-immutable/no-expression-statement": "error",
            "ts-immutable/no-try": "error"
        }
    }
]);

// Conditionally loaded TypeScript but only if it is avaliable.
var ts = (() => {
    try {
        return require("typescript");
    }
    catch (error) {
        return undefined;
    }
})();

/**
 * @file Functions that typeguard the given node/type.
 */
/*
 * Node type guards.
 */
function isArrayExpression(node) {
    return node.type === AST_NODE_TYPES.ArrayExpression;
}
function isAssignmentExpression(node) {
    return node.type === AST_NODE_TYPES.AssignmentExpression;
}
function isAssignmentPattern(node) {
    return node.type === AST_NODE_TYPES.AssignmentPattern;
}
function isBlockStatement(node) {
    return node.type === AST_NODE_TYPES.BlockStatement;
}
function isCallExpression(node) {
    return node.type === AST_NODE_TYPES.CallExpression;
}
/**
 * Is the given node a class node?
 *
 * It doesn't matter what type of class.
 */
function isClassLike(node) {
    return (node.type === AST_NODE_TYPES.ClassDeclaration ||
        node.type === AST_NODE_TYPES.ClassExpression);
}
/**
 * Is the given node a function node?
 *
 * It doesn't matter what type of function.
 */
function isFunctionLike(node) {
    return (node.type === AST_NODE_TYPES.FunctionDeclaration ||
        node.type === AST_NODE_TYPES.FunctionExpression ||
        node.type === AST_NODE_TYPES.ArrowFunctionExpression);
}
function isIdentifier(node) {
    return node.type === AST_NODE_TYPES.Identifier;
}
function isIfStatement(node) {
    return node.type === AST_NODE_TYPES.IfStatement;
}
function isMemberExpression(node) {
    return node.type === AST_NODE_TYPES.MemberExpression;
}
function isMethodDefinition(node) {
    return node.type === AST_NODE_TYPES.MethodDefinition;
}
function isNewExpression(node) {
    return node.type === AST_NODE_TYPES.NewExpression;
}
function isProperty(node) {
    return node.type === AST_NODE_TYPES.Property;
}
function isRestElement(node) {
    return node.type === AST_NODE_TYPES.RestElement;
}
function isReturnStatement(node) {
    return node.type === AST_NODE_TYPES.ReturnStatement;
}
function isTSArrayType(node) {
    return node.type === AST_NODE_TYPES.TSArrayType;
}
function isTSIndexSignature(node) {
    return node.type === AST_NODE_TYPES.TSIndexSignature;
}
function isTSInterfaceBody(node) {
    return node.type === AST_NODE_TYPES.TSInterfaceBody;
}
function isTSNullKeyword(node) {
    return node.type === AST_NODE_TYPES.TSNullKeyword;
}
function isTSParameterProperty(node) {
    return node.type === AST_NODE_TYPES.TSParameterProperty;
}
function isTSPropertySignature(node) {
    return node.type === AST_NODE_TYPES.TSPropertySignature;
}
function isTSTupleType(node) {
    return node.type === AST_NODE_TYPES.TSTupleType;
}
function isTSTypeAliasDeclaration(node) {
    return node.type === AST_NODE_TYPES.TSTypeAliasDeclaration;
}
function isTSTypeOperator(node) {
    return node.type === AST_NODE_TYPES.TSTypeOperator;
}
function isTSUndefinedKeyword(node) {
    return node.type === AST_NODE_TYPES.TSUndefinedKeyword;
}
function isTSVoidKeyword(node) {
    return node.type === AST_NODE_TYPES.TSVoidKeyword;
}
function isVariableDeclaration(node) {
    return node.type === AST_NODE_TYPES.VariableDeclaration;
}
function isVariableDeclarator(node) {
    return node.type === AST_NODE_TYPES.VariableDeclarator;
}
/*
 * TS types type guards.
 */
function isUnionType(type) {
    return ts !== undefined && type.flags === ts.TypeFlags.Union;
}
function isArrayType(type, assumeType = false, node = null) {
    return assumeType === true
        ? node !== null
        : type !== null &&
            ((type.symbol && type.symbol.name === "Array") ||
                (isUnionType(type) &&
                    type.types.some(t => isArrayType(t, false, null))));
}
function isArrayConstructorType(type, assumeType = false, node = null) {
    return assumeType === true
        ? node !== null && isIdentifier(node) && node.name === "Array"
        : type !== null &&
            ((type.symbol && type.symbol.name === "ArrayConstructor") ||
                (isUnionType(type) &&
                    type.types.some(t => isArrayConstructorType(t, false, null))));
}
function isObjectConstructorType(type, assumeType = false, node = null) {
    return assumeType === true
        ? node !== null && isIdentifier(node) && node.name === "Object"
        : type !== null &&
            ((type.symbol && type.symbol.name === "ObjectConstructor") ||
                (isUnionType(type) &&
                    type.types.some(t => isObjectConstructorType(t, false, null))));
}

/**
 * Return the parent that meets the given check criteria.
 */
function getParentOfType(checker, node) {
    return checker(node)
        ? node
        : node.parent == undefined
            ? null
            : getParentOfType(checker, node.parent);
}
/**
 * Test if the given node is in a function.
 */
function inFunction(node) {
    return getParentOfType(isFunctionLike, node) !== null;
}
/**
 * Test if the given node is in a class.
 */
function inClass(node) {
    return getParentOfType(isClassLike, node) !== null;
}
/**
 * Test if the given node is in a TS Property Signature.
 */
function inInterface(node) {
    return getParentOfType(isTSInterfaceBody, node) !== null;
}
/**
 * Test if the given node is in a Constructor.
 */
function inConstructor(node) {
    const methodDefinition = getParentOfType(isMethodDefinition, node);
    return (methodDefinition !== null &&
        isIdentifier(methodDefinition.key) &&
        methodDefinition.key.name === "constructor");
}
/**
 * Is the given node in the return type.
 */
function isInReturnType(node) {
    return (getParentOfType((n) => {
        return (n.parent != undefined &&
            isFunctionLike(n.parent) &&
            n.parent.returnType === n);
    }, node) !== null);
}
/**
 * Is the given identifier a property of an object?
 */
function isPropertyAccess(node) {
    return (node.parent !== undefined &&
        isMemberExpression(node.parent) &&
        node.parent.property === node);
}
/**
 * Is the given identifier a property name?
 */
function isPropertyName(node) {
    return (node.parent !== undefined &&
        isProperty(node.parent) &&
        node.parent.key === node);
}

// Polyfill.
const ignoreLocalOptionSchema = {
    type: "object",
    properties: {
        ignoreLocal: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
const ignorePatternOptionSchema = {
    type: "object",
    properties: {
        ignorePattern: {
            type: ["string", "array"],
            items: {
                type: "string"
            }
        }
    },
    additionalProperties: false
};
const ignoreAccessorPatternOptionSchema = {
    type: "object",
    properties: {
        ignoreAccessorPattern: {
            type: ["string", "array"],
            items: {
                type: "string"
            }
        }
    },
    additionalProperties: false
};
const ignoreReturnTypeOptionSchema = {
    type: "object",
    properties: {
        ignoreReturnType: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
const ignoreClassOptionSchema = {
    type: "object",
    properties: {
        ignoreClass: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
const ignoreInterfaceOptionSchema = {
    type: "object",
    properties: {
        ignoreInterface: {
            type: "boolean"
        }
    },
    additionalProperties: false
};
/**
 * Recursive callback of `getNodeText`.
 *
 * This function not be called from anywhere else.
 */
function _getNodeText(node, context) {
    return isIdentifier(node)
        ? node.name
        : isMemberExpression(node)
            ? `${_getNodeText(node.object, context)}.${_getNodeText(node.property, context)}`
            : context.getSourceCode().getText(node);
}
/**
 * Get the text of the given node.
 */
function getNodeText(node, context) {
    return isAssignmentExpression(node)
        ? getNodeText(node.left, context)
        : isCallExpression(node)
            ? getNodeText(node.callee, context)
            : isMemberExpression(node)
                ? _getNodeText(node.object, context)
                : isVariableDeclarator(node) || isTSTypeAliasDeclaration(node)
                    ? _getNodeText(node.id, context)
                    : isTSPropertySignature(node)
                        ? _getNodeText(node.key, context)
                        : _getNodeText(node, context);
}
/**
 * Get all the important bits of texts from the given node.
 */
function getNodeTexts(node, context) {
    return (isVariableDeclaration(node)
        ? node.declarations.flatMap(declarator => getNodeText(declarator, context))
        : [getNodeText(node, context)]).filter(name => name !== undefined);
}
/**
 * Should the given text be ignore?
 *
 * Test using the given pattern(s).
 */
function shouldIgnoreViaPattern(text, ignorePattern) {
    const patterns = Array.isArray(ignorePattern)
        ? ignorePattern
        : [ignorePattern];
    // One or more patterns match?
    return patterns.some(pattern => new RegExp(pattern).test(text));
}
/**
 * Recursive callback of `shouldIgnoreViaAccessorPattern`.
 *
 * This function not be called from anywhere else.
 *
 * Does the given text match the given pattern.
 */
function accessorPatternMatch([pattern, ...remainingPatternParts], textParts, allowExtra = false) {
    return pattern === undefined
        ? allowExtra || textParts.length === 0
        : // Match any depth (including 0)?
            pattern === "**"
                ? textParts.length === 0
                    ? accessorPatternMatch(remainingPatternParts, [], allowExtra)
                    : Array.from({ length: textParts.length })
                        .map((_element, index) => index)
                        .some(offset => accessorPatternMatch(remainingPatternParts, textParts.slice(offset), true))
                : // Match anything?
                    pattern === "*"
                        ? textParts.length > 0 &&
                            accessorPatternMatch(remainingPatternParts, textParts.slice(1), allowExtra)
                        : // Text matches pattern?
                            new RegExp("^" + escapeRegExp(pattern).replace(/\\\*/g, ".*") + "$").test(textParts[0]) &&
                                accessorPatternMatch(remainingPatternParts, textParts.slice(1), allowExtra);
}
/**
 * Should the given text be ignore?
 *
 * Test using the given accessor pattern(s).
 */
function shouldIgnoreViaAccessorPattern(text, ignorePattern) {
    const patterns = Array.isArray(ignorePattern)
        ? ignorePattern
        : [ignorePattern];
    // One or more patterns match?
    return patterns.some(pattern => accessorPatternMatch(pattern.split("."), text.split(".")));
}
/**
 * Should the given node be ignored?
 */
function shouldIgnore(node, context, options) {
    return (
    // Ignore if in a function and ignoreLocal is set.
    (Boolean(options.ignoreLocal) && inFunction(node)) ||
        // Ignore if in a class and ignoreClass is set.
        (Boolean(options.ignoreClass) && inClass(node)) ||
        // Ignore if in an interface and ignoreInterface is set.
        (Boolean(options.ignoreInterface) && inInterface(node)) ||
        ((texts) => texts.length > 0
            ? // Ignore if ignorePattern is set and a pattern matches.
                (options.ignorePattern !== undefined &&
                    texts.every(text => shouldIgnoreViaPattern(text, options.ignorePattern))) ||
                    // Ignore if ignoreAccessorPattern is set and an accessor pattern matches.
                    (options.ignoreAccessorPattern !== undefined &&
                        texts.every(text => shouldIgnoreViaAccessorPattern(text, options.ignoreAccessorPattern)))
            : false)(getNodeTexts(node, context)));
}

const version = "0.2.1";

/**
 * Create a rule.
 */
function createRule(name, meta, defaultOptions, create) {
    return ESLintUtils.RuleCreator(name => `https://github.com/jonaskello/eslint-plugin-ts-immutable/blob/v${version}/docs/rules/${name}.md`)({
        name,
        meta,
        defaultOptions: [defaultOptions],
        create: (c, [o]) => create(c, o)
        /* eslint-disable-next-line @typescript-eslint/no-explicit-any */
    });
}
// This function can't be functional as it needs to interact with 3rd-party
// libraries that aren't functional.
/* eslint-disable ts-immutable/no-return-void, ts-immutable/no-conditional-statement, ts-immutable/no-expression-statement */
/**
 * Create a function that processes common options and then runs the given
 * check.
 */
function checkNode(check, context, options) {
    return (node) => {
        if (!options || !shouldIgnore(node, context, options)) {
            const result = check(node, context, options);
            result.descriptors.forEach(descriptor => result.context.report(descriptor));
        }
    };
}
/* eslint-enable ts-immutable/no-return-void, ts-immutable/no-conditional-statement, ts-immutable/no-expression-statement */
/**
 * Get the type of the the given node.
 */
function getTypeOfNode(node, context) {
    const parserServices = context.parserServices;
    return parserServices === undefined ||
        parserServices.program === undefined ||
        parserServices.esTreeNodeToTSNodeMap === undefined
        ? null
        : parserServices.program
            .getTypeChecker()
            .getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node));
}

// The name of this rule.
const name = "functional-parameters";
// The schema for the rule options.
const schema = [
    all([
        ignorePatternOptionSchema,
        {
            type: "object",
            properties: {
                allowRestParameter: {
                    type: "boolean"
                },
                allowArgumentsKeyword: {
                    type: "boolean"
                },
                enforceParameterCount: {
                    oneOf: [
                        {
                            type: "boolean",
                            enum: [false]
                        },
                        {
                            type: "string",
                            enum: ["atLeastOne", "exactlyOne"]
                        }
                    ]
                }
            },
            additionalProperties: false
        }
    ])
];
// The default options for the rule.
const defaultOptions = {
    allowRestParameter: false,
    allowArgumentsKeyword: false,
    enforceParameterCount: "atLeastOne"
};
// The possible error messages.
const errorMessages = {
    restParam: "Unexpected rest parameter. Use a regular parameter of type array instead.",
    arguments: "Unexpected use of `arguments`. Use regular function arguments instead.",
    paramCountAtLeastOne: "Functions must have at least one parameter.",
    paramCountExactlyOne: "Functions must have exactly one parameter."
};
// The meta data for this rule.
const meta = {
    type: "suggestion",
    docs: {
        description: "Enforce functional parameters.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages,
    schema
};
/**
 * Get the rest parameter violations.
 */
function getRestParamViolations(allowRestParameter, node) {
    return !allowRestParameter &&
        node.params.length > 0 &&
        isRestElement(node.params[node.params.length - 1])
        ? [
            {
                node: node.params[node.params.length - 1],
                messageId: "restParam"
            }
        ]
        : [];
}
/**
 * Get the parameter count violations.
 */
function getParamCountViolations(enforceParameterCount, node) {
    return enforceParameterCount === "atLeastOne" && node.params.length < 1
        ? [
            {
                node: node,
                messageId: "paramCountAtLeastOne"
            }
        ]
        : enforceParameterCount === "exactlyOne" && node.params.length !== 1
            ? [
                {
                    node: node,
                    messageId: "paramCountExactlyOne"
                }
            ]
            : [];
}
/**
 * Check if the given function node has a reset parameter this rule.
 */
function checkFunction(node, context, options) {
    return {
        context,
        descriptors: [
            ...getRestParamViolations(options.allowRestParameter, node),
            ...getParamCountViolations(options.enforceParameterCount, node)
        ]
    };
}
/**
 * Check if the given identifier is for the "arguments" keyword.
 */
function checkIdentifier(node, context, options) {
    return {
        context,
        descriptors: !options.allowArgumentsKeyword &&
            node.name === "arguments" &&
            !isPropertyName(node) &&
            !isPropertyAccess(node)
            ? [
                {
                    node,
                    messageId: "arguments"
                }
            ]
            : []
    };
}
// Create the rule.
const rule = createRule(name, meta, defaultOptions, (context, options) => {
    const _checkFunction = checkNode(checkFunction, context, options);
    const _checkIdentifier = checkNode(checkIdentifier, context, options);
    return {
        FunctionDeclaration: _checkFunction,
        FunctionExpression: _checkFunction,
        ArrowFunctionExpression: _checkFunction,
        Identifier: _checkIdentifier
    };
});

/**
 * Returns a function that checks if the given value is the same as the expected
 * value.
 */
function isExpected(expected) {
    return actual => actual === expected;
}

// The name of this rule.
const name$1 = "immutable-data";
// The schema for the rule options.
const schema$1 = [
    all([
        ignorePatternOptionSchema,
        ignoreAccessorPatternOptionSchema,
        {
            type: "object",
            properties: {
                assumeTypes: {
                    oneOf: [
                        {
                            type: "boolean"
                        },
                        {
                            type: "object",
                            properties: {
                                forArrays: {
                                    type: "boolean"
                                },
                                forObjects: {
                                    type: "boolean"
                                }
                            },
                            additionalProperties: false
                        }
                    ]
                }
            },
            additionalProperties: false
        }
    ])
];
// The default options for the rule.
const defaultOptions$1 = {
    assumeTypes: true
};
// The possible error messages.
const errorMessages$1 = {
    generic: "Modifying an existing object/array is not allowed.",
    object: "Modifying properties of existing object not allowed.",
    array: "Modifying an array is not allowed."
};
// The meta data for this rule.
const meta$1 = {
    type: "suggestion",
    docs: {
        description: "Enforce treating data as immutable.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$1,
    schema: schema$1
};
/**
 * Array methods that mutate an array.
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/prototype#Methods#Mutator_methods
 */
const arrayMutatorMethods = [
    "copyWithin",
    "fill",
    "pop",
    "push",
    "reverse",
    "shift",
    "sort",
    "splice",
    "unshift"
];
/**
 * Array methods that return a new object (or array) without mutating the original.
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/prototype#Methods#Accessor_methods
 * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array/prototype#Iteration_methods
 */
const arrayNewObjectReturningMethods = [
    "concat",
    "slice",
    "filter",
    "map",
    "reduce",
    "reduceRight"
];
/**
 * Array constructor functions that create a new array.
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Array#Methods
 */
const arrayConstructorFunctions = ["from", "of"];
/**
 * Object constructor functions that mutate an object.
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Object#Methods_of_the_Object_constructor
 */
const objectConstructorMutatorFunctions = [
    "assign",
    "defineProperties",
    "defineProperty",
    "setPrototypeOf"
];
/**
 * Check if the given assignment expression violates this rule.
 */
function checkAssignmentExpression(node, context) {
    return {
        context,
        descriptors: isMemberExpression(node.left) &&
            // Ignore if in a constructor - allow for field initialization.
            !inConstructor(node)
            ? [{ node, messageId: "generic" }]
            : []
    };
}
/**
 * Check if the given node violates this rule.
 */
function checkUnaryExpression(node, context) {
    return {
        context,
        descriptors: node.operator === "delete" && isMemberExpression(node.argument)
            ? [{ node, messageId: "generic" }]
            : []
    };
}
/**
 * Check if the given node violates this rule.
 */
function checkUpdateExpression(node, context) {
    return {
        context,
        descriptors: isMemberExpression(node.argument)
            ? [{ node, messageId: "generic" }]
            : []
    };
}
/**
 * Check if the given the given MemberExpression is part of a chain and
 * immediately follows a method/function call that returns a new array.
 *
 * If this is the case, then the given MemberExpression is allowed to be
 * a mutator method call.
 */
function isInChainCallAndFollowsNew(node, context, assumeArrayTypes) {
    return (
    // Check for: [0, 1, 2]
    isArrayExpression(node.object) ||
        // Check for: new Array()
        ((isNewExpression(node.object) &&
            isArrayConstructorType(getTypeOfNode(node.object.callee, context), assumeArrayTypes, node.object.callee)) ||
            (isCallExpression(node.object) &&
                isMemberExpression(node.object.callee) &&
                isIdentifier(node.object.callee.property) &&
                // Check for: Array.from(iterable)
                ((arrayConstructorFunctions.some(isExpected(node.object.callee.property.name)) &&
                    isArrayConstructorType(getTypeOfNode(node.object.callee.object, context), assumeArrayTypes, node.object.callee.object)) ||
                    // Check for: array.slice(0)
                    arrayNewObjectReturningMethods.some(isExpected(node.object.callee.property.name))))));
}
/**
 * Check if the given node violates this rule.
 */
function checkCallExpression(node, context, options) {
    const assumeTypesForArrays = options.assumeTypes === true ||
        (options.assumeTypes !== false && Boolean(options.assumeTypes.forArrays));
    const assumeTypesForObjects = options.assumeTypes === true ||
        (options.assumeTypes !== false && Boolean(options.assumeTypes.forObjects));
    return {
        context,
        descriptors: 
        // Potential object mutation?
        isMemberExpression(node.callee) && isIdentifier(node.callee.property)
            ? // Potential array mutation?
                arrayMutatorMethods.some(m => m ===
                    node.callee
                        .property.name) &&
                    !isInChainCallAndFollowsNew(node.callee, context, assumeTypesForArrays) &&
                    isArrayType(getTypeOfNode(node.callee.object, context), assumeTypesForArrays, node.callee.object)
                    ? [{ node, messageId: "array" }]
                    : // Potential non-array object mutation (ex. Object.assign on identifier)?
                        objectConstructorMutatorFunctions.some(m => m ===
                            node.callee
                                .property.name) &&
                            node.arguments.length >= 2 &&
                            (isIdentifier(node.arguments[0]) ||
                                isMemberExpression(node.arguments[0])) &&
                            isObjectConstructorType(getTypeOfNode(node.callee.object, context), assumeTypesForObjects, node.callee.object)
                            ? [{ node, messageId: "object" }]
                            : []
            : []
    };
}
// Create the rule.
const rule$1 = createRule(name$1, meta$1, defaultOptions$1, (context, options) => {
    const _checkAssignmentExpression = checkNode(checkAssignmentExpression, context, options);
    const _checkUnaryExpression = checkNode(checkUnaryExpression, context, options);
    const _checkUpdateExpression = checkNode(checkUpdateExpression, context, options);
    // This functionality is only avaliable if the parser services are
    // avaliable.
    const _checkCallExpression = checkNode(checkCallExpression, context, options);
    return {
        AssignmentExpression: _checkAssignmentExpression,
        UnaryExpression: _checkUnaryExpression,
        UpdateExpression: _checkUpdateExpression,
        CallExpression: _checkCallExpression
    };
});

// The name of this rule.
const name$2 = "no-class";
// The schema for the rule options.
const schema$2 = [];
// The default options for the rule.
const defaultOptions$2 = {};
// The possible error messages.
const errorMessages$2 = {
    generic: "Unexpected class, use functions not classes."
};
// The meta data for this rule.
const meta$2 = {
    type: "suggestion",
    docs: {
        description: "Disallow classes.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$2,
    schema: schema$2
};
/**
 * Check if the given class node violates this rule.
 */
function checkClass(node, context) {
    // All class nodes violate this rule.
    return { context, descriptors: [{ node, messageId: "generic" }] };
}
// Create the rule.
const rule$2 = createRule(name$2, meta$2, defaultOptions$2, (context, options) => {
    const _checkClass = checkNode(checkClass, context, options);
    return {
        ClassDeclaration: _checkClass,
        ClassExpression: _checkClass
    };
});

// The name of this rule.
const name$3 = "no-return-void";
// The schema for the rule options.
const schema$3 = [
    {
        type: "object",
        properties: {
            allowNull: {
                type: "boolean"
            },
            allowUndefined: {
                type: "boolean"
            }
        },
        additionalProperties: false
    }
];
// The default options for the rule.
const defaultOptions$3 = {
    allowNull: true,
    allowUndefined: true
};
// The possible error messages.
const errorMessages$3 = {
    generic: "Function must return a value."
};
// The meta data for this rule.
const meta$3 = {
    type: "suggestion",
    docs: {
        description: "Disallow functions that don't return anything.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$3,
    schema: schema$3
};
/**
 * Check if the given function node violates this rule.
 */
function checkFunction$1(node, context, options) {
    return {
        context,
        descriptors: node.returnType !== undefined &&
            (isTSVoidKeyword(node.returnType.typeAnnotation) ||
                (!options.allowNull &&
                    isTSNullKeyword(node.returnType.typeAnnotation)) ||
                (!options.allowUndefined &&
                    isTSUndefinedKeyword(node.returnType.typeAnnotation)))
            ? [{ node: node.returnType, messageId: "generic" }]
            : []
    };
}
// Create the rule.
const rule$3 = createRule(name$3, meta$3, defaultOptions$3, (context, options) => {
    const _checkFunction = checkNode(checkFunction$1, context, options);
    return {
        FunctionDeclaration: _checkFunction,
        FunctionExpression: _checkFunction,
        ArrowFunctionExpression: _checkFunction,
        TSFunctionType: _checkFunction
    };
});

// The name of this rule.
const name$4 = "no-conditional-statement";
// The schema for the rule options.
const schema$4 = [
    {
        type: "object",
        properties: {
            allowReturningBranches: {
                oneOf: [
                    {
                        type: "boolean"
                    },
                    {
                        type: "string",
                        enum: ["ifExhaustive"]
                    }
                ]
            }
        },
        additionalProperties: false
    }
];
// The default options for the rule.
const defaultOptions$4 = { allowReturningBranches: false };
// The possible error messages.
const errorMessages$4 = {
    incompleteBranch: "Incomplete branch, every branch in a conditional statement must contain a return statement.",
    incompleteIf: "Incomplete if, it must have an else statement and every branch must contain a return statement.",
    incompleteSwitch: "Incomplete switch, it must have an default case and every case must contain a return statement.",
    unexpectedIf: "Unexpected if, use a conditional expression (ternary operator) instead.",
    unexpectedSwitch: "Unexpected switch, use a conditional expression (ternary operator) instead."
};
// The meta data for this rule.
const meta$4 = {
    type: "suggestion",
    docs: {
        description: "Disallow conditional statements.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$4,
    schema: schema$4
};
/**
 * Get all of the violations in the given if statement assuming if statements
 * are allowed.
 */
function getIfBranchViolations(node) {
    const nodes = [node.consequent, node.alternate].reduce((carry, branch) => branch === null ||
        isReturnStatement(branch) ||
        (isBlockStatement(branch) &&
            branch.body.some(statement => isReturnStatement(statement) ||
                // Another instance of this rule will check nested if statements.
                isIfStatement(statement))) ||
        isIfStatement(branch)
        ? carry
        : [...carry, branch], []);
    return nodes.flatMap(node => [{ node, messageId: "incompleteBranch" }]);
}
/**
 * Get all of the violations in the given switch statement assuming switch
 * statements are allowed.
 */
function getSwitchViolations(node) {
    const nodes = node.cases.reduce((carry, branch) => branch.consequent.length === 0 ||
        branch.consequent.some(isReturnStatement) ||
        (branch.consequent.every(isBlockStatement) &&
            branch.consequent[branch.consequent.length - 1].body.some(isReturnStatement))
        ? carry
        : [...carry, branch], []);
    return nodes.flatMap(node => [{ node, messageId: "incompleteBranch" }]);
}
/**
 * Does the given if statement violate this rule if it must be exhaustive.
 */
function isExhaustiveIfViolation(node) {
    return node.alternate === null;
}
/**
 * Does the given switch statement violate this rule if it must be exhaustive.
 */
function isExhaustiveSwitchViolation(node) {
    return (
    // No cases defined.
    node.cases.length === 0 ||
        // No default case defined.
        node.cases.every(c => c.test !== null));
}
/**
 * Check if the given IfStatement violates this rule.
 */
function checkIfStatement(node, context, options) {
    return {
        context,
        descriptors: options.allowReturningBranches
            ? options.allowReturningBranches === "ifExhaustive"
                ? isExhaustiveIfViolation(node)
                    ? [{ node, messageId: "incompleteIf" }]
                    : getIfBranchViolations(node)
                : getIfBranchViolations(node)
            : [{ node, messageId: "unexpectedIf" }]
    };
}
/**
 * Check if the given SwitchStatement violates this rule.
 */
function checkSwitchStatement(node, context, options) {
    return {
        context,
        descriptors: options.allowReturningBranches
            ? options.allowReturningBranches === "ifExhaustive"
                ? isExhaustiveSwitchViolation(node)
                    ? [{ node, messageId: "incompleteSwitch" }]
                    : getSwitchViolations(node)
                : getSwitchViolations(node)
            : [{ node, messageId: "unexpectedSwitch" }]
    };
}
// Create the rule.
const rule$4 = createRule(name$4, meta$4, defaultOptions$4, (context, options) => {
    const _checkIfStatement = checkNode(checkIfStatement, context, options);
    const _checkSwitchStatement = checkNode(checkSwitchStatement, context, options);
    return {
        IfStatement: _checkIfStatement,
        SwitchStatement: _checkSwitchStatement
    };
});

// The name of this rule.
const name$5 = "no-expression-statement";
// The schema for the rule options.
const schema$5 = [ignorePatternOptionSchema];
// The default options for the rule.
const defaultOptions$5 = {};
// The possible error messages.
const errorMessages$5 = {
    generic: "Using expressions to cause side-effects not allowed."
};
// The meta data for this rule.
const meta$5 = {
    type: "suggestion",
    docs: {
        description: "Disallow expression statements.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$5,
    schema: schema$5
};
/**
 * Check if the given ExpressionStatement violates this rule.
 */
function checkExpressionStatement(node, context) {
    // All expression statements violate this rule.
    return { context, descriptors: [{ node, messageId: "generic" }] };
}
// Create the rule.
const rule$5 = createRule(name$5, meta$5, defaultOptions$5, (context, options) => {
    const _checkExpressionStatement = checkNode(checkExpressionStatement, context, options);
    return {
        ExpressionStatement: _checkExpressionStatement
    };
});

// The name of this rule.
const name$6 = "no-let";
// The schema for the rule options.
const schema$6 = [
    all([ignoreLocalOptionSchema, ignorePatternOptionSchema])
];
// The default options for the rule.
const defaultOptions$6 = {
    ignoreLocal: false
};
// The possible error messages.
const errorMessages$6 = {
    generic: "Unexpected let, use const instead."
};
// The meta data for this rule.
const meta$6 = {
    type: "suggestion",
    docs: {
        description: "Disallow mutable variables.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$6,
    fixable: "code",
    schema: schema$6
};
/**
 * Check if the given VariableDeclaration violates this rule.
 */
function checkVariableDeclaration(node, context) {
    return {
        context,
        descriptors: node.kind === "let" ? [{ node, messageId: "generic" }] : []
    };
}
// Create the rule.
const rule$6 = createRule(name$6, meta$6, defaultOptions$6, (context, options) => {
    const _checkVariableDeclaration = checkNode(checkVariableDeclaration, context, options);
    return {
        VariableDeclaration: _checkVariableDeclaration
    };
});

// The name of this rule.
const name$7 = "no-loop-statement";
// The schema for the rule options.
const schema$7 = [];
// The default options for the rule.
const defaultOptions$7 = {};
// The possible error messages.
const errorMessages$7 = {
    generic: "Unexpected loop, use map or reduce instead."
};
// The meta data for this rule.
const meta$7 = {
    type: "suggestion",
    docs: {
        description: "Disallow imperative loops.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$7,
    schema: schema$7
};
/**
 * Check if the given loop violates this rule.
 */
function checkLoop(node, context) {
    // All loops violate this rule.
    return { context, descriptors: [{ node, messageId: "generic" }] };
}
// Create the rule.
const rule$7 = createRule(name$7, meta$7, defaultOptions$7, (context, options) => {
    const _checkLoop = checkNode(checkLoop, context, options);
    return {
        ForStatement: _checkLoop,
        ForInStatement: _checkLoop,
        ForOfStatement: _checkLoop,
        WhileStatement: _checkLoop,
        DoWhileStatement: _checkLoop
    };
});

// The name of this rule.
const name$8 = "no-method-signature";
// The schema for the rule options.
const schema$8 = [];
// The default options for the rule.
const defaultOptions$8 = {};
// The possible error messages.
const errorMessages$8 = {
    generic: "Method signature is mutable, use property signature with readonly modifier instead."
};
// The meta data for this rule.
const meta$8 = {
    type: "suggestion",
    docs: {
        description: "Prefer property signatures with readonly modifiers over method signatures.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$8,
    schema: schema$8
};
/**
 * Check if the given TSMethodSignature violates this rule.
 */
function checkTSMethodSignature(node, context) {
    // All TS method signatures violate this rule.
    return { context, descriptors: [{ node, messageId: "generic" }] };
}
// Create the rule.
const rule$8 = createRule(name$8, meta$8, defaultOptions$8, (context, options) => {
    const _checkTSMethodSignature = checkNode(checkTSMethodSignature, context, options);
    return {
        TSMethodSignature: _checkTSMethodSignature
    };
});

// The name of this rule.
const name$9 = "no-mixed-interface";
// The schema for the rule options.
const schema$9 = [];
// The default options for the rule.
const defaultOptions$9 = {};
// The possible error messages.
const errorMessages$9 = {
    generic: "Only the same kind of members allowed in interfaces."
};
// The meta data for this rule.
const meta$9 = {
    type: "suggestion",
    docs: {
        description: "Restrict interfaces so that only members of the same kind of are allowed in them.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$9,
    schema: schema$9
};
/**
 * Check if the given TSInterfaceDeclaration violates this rule.
 */
function checkTSInterfaceDeclaration(node, context) {
    return {
        context,
        descriptors: node.body.body.reduce((carry, member) => {
            const memberType = member.type;
            const memberTypeAnnotation = isTSPropertySignature(member) && member.typeAnnotation !== undefined
                ? member.typeAnnotation.typeAnnotation.type
                : undefined;
            return {
                prevMemberType: memberType,
                prevMemberTypeAnnotation: memberTypeAnnotation,
                violations: 
                // Not the first property in the interface.
                carry.prevMemberType !== undefined &&
                    // And different property type to previous property.
                    (carry.prevMemberType !== memberType ||
                        // Or annotationed with a different type annotation.
                        (carry.prevMemberTypeAnnotation !== memberTypeAnnotation &&
                            // Where one of the properties is a annotationed as a function.
                            (carry.prevMemberTypeAnnotation ===
                                AST_NODE_TYPES$1.TSFunctionType ||
                                memberTypeAnnotation === AST_NODE_TYPES$1.TSFunctionType)))
                    ? [...carry.violations, { node: member, messageId: "generic" }]
                    : carry.violations
            };
        }, {
            prevMemberType: undefined,
            prevMemberTypeAnnotation: undefined,
            violations: []
        }).violations
    };
}
// Create the rule.
const rule$9 = createRule(name$9, meta$9, defaultOptions$9, (context, options) => {
    const _checkTSInterfaceDeclaration = checkNode(checkTSInterfaceDeclaration, context, options);
    return {
        TSInterfaceDeclaration: _checkTSInterfaceDeclaration
    };
});

// The name of this rule.
const name$a = "no-reject";
// The schema for the rule options.
const schema$a = [];
// The default options for the rule.
const defaultOptions$a = {};
// The possible error messages.
const errorMessages$a = {
    generic: "Unexpected reject, return an error instead."
};
// The meta data for this rule.
const meta$a = {
    type: "suggestion",
    docs: {
        description: "Disallow try-catch[-finally] and try-finally patterns.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$a,
    schema: schema$a
};
/**
 * Check if the given CallExpression violates this rule.
 */
function checkCallExpression$1(node, context) {
    return {
        context,
        descriptors: isMemberExpression(node.callee) &&
            isIdentifier(node.callee.object) &&
            isIdentifier(node.callee.property) &&
            node.callee.object.name === "Promise" &&
            node.callee.property.name === "reject"
            ? [{ node, messageId: "generic" }]
            : []
    };
}
// Create the rule.
const rule$a = createRule(name$a, meta$a, defaultOptions$a, (context, options) => {
    const _checkCallExpression = checkNode(checkCallExpression$1, context, options);
    return {
        CallExpression: _checkCallExpression
    };
});

// The name of this rule.
const name$b = "no-this";
// The schema for the rule options.
const schema$b = [];
// The default options for the rule.
const defaultOptions$b = {};
// The possible error messages.
const errorMessages$b = {
    generic: "Unexpected this, use functions not classes."
};
// The meta data for this rule.
const meta$b = {
    type: "suggestion",
    docs: {
        description: "Disallow this access.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$b,
    schema: schema$b
};
/**
 * Check if the given ThisExpression violates this rule.
 */
function checkThisExpression(node, context) {
    // All throw statements violate this rule.
    return { context, descriptors: [{ node, messageId: "generic" }] };
}
// Create the rule.
const rule$b = createRule(name$b, meta$b, defaultOptions$b, (context, options) => {
    const _checkThisExpression = checkNode(checkThisExpression, context, options);
    return {
        ThisExpression: _checkThisExpression
    };
});

// The name of this rule.
const name$c = "no-throw";
// The schema for the rule options.
const schema$c = [];
// The default options for the rule.
const defaultOptions$c = {};
// The possible error messages.
const errorMessages$c = {
    generic: "Unexpected throw, throwing exceptions is not functional."
};
// The meta data for this rule.
const meta$c = {
    type: "suggestion",
    docs: {
        description: "Disallow throwing exceptions.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$c,
    schema: schema$c
};
/**
 * Check if the given ThrowStatement violates this rule.
 */
function checkThrowStatement(node, context) {
    // All throw statements violate this rule.
    return { context, descriptors: [{ node, messageId: "generic" }] };
}
// Create the rule.
const rule$c = createRule(name$c, meta$c, defaultOptions$c, (context, options) => {
    const _checkThrowStatement = checkNode(checkThrowStatement, context, options);
    return {
        ThrowStatement: _checkThrowStatement
    };
});

// The name of this rule.
const name$d = "no-try";
// The schema for the rule options.
const schema$d = [];
// The default options for the rule.
const defaultOptions$d = {};
// The possible error messages.
const errorMessages$d = {
    generic: "Unexpected try, the try-catch[-finally] and try-finally patterns are not functional."
};
// The meta data for this rule.
const meta$d = {
    type: "suggestion",
    docs: {
        description: "Disallow try-catch[-finally] and try-finally patterns.",
        category: "Best Practices",
        recommended: false
    },
    messages: errorMessages$d,
    schema: schema$d
};
/**
 * Check if the given TryStatement violates this rule.
 */
function checkTryStatement(node, context) {
    // All try statements violate this rule.
    return { context, descriptors: [{ node, messageId: "generic" }] };
}
// Create the rule.
const rule$d = createRule(name$d, meta$d, defaultOptions$d, (context, options) => {
    const _checkTryStatement = checkNode(checkTryStatement, context, options);
    return {
        TryStatement: _checkTryStatement
    };
});

// Polyfill.
// The name of this rule.
const name$e = "prefer-readonly-types";
// The schema for the rule options.
const schema$e = [
    all([
        ignoreLocalOptionSchema,
        ignorePatternOptionSchema,
        ignoreClassOptionSchema,
        ignoreInterfaceOptionSchema,
        ignoreReturnTypeOptionSchema,
        {
            type: "object",
            properties: {
                checkImplicit: {
                    type: "boolean"
                }
            },
            additionalProperties: false
        }
    ])
];
// The default options for the rule.
const defaultOptions$e = {
    checkImplicit: false,
    ignoreClass: false,
    ignoreInterface: false,
    ignoreLocal: false,
    ignoreReturnType: false
};
// The possible error messages.
const errorMessages$e = {
    array: "Only readonly arrays allowed.",
    implicit: "Implicitly a mutable array. Only readonly arrays allowed.",
    property: "A readonly modifier is required.",
    tuple: "Only readonly tuples allowed.",
    type: "Only readonly types allowed."
};
// The meta data for this rule.
const meta$e = {
    type: "suggestion",
    docs: {
        description: "Prefer readonly array over mutable arrays.",
        category: "Best Practices",
        recommended: "error"
    },
    messages: errorMessages$e,
    fixable: "code",
    schema: schema$e
};
const mutableToImmutableTypes = new Map([["Array", "ReadonlyArray"], ["Map", "ReadonlyMap"], ["Set", "ReadonlySet"]]);
/**
 * Check if the given ArrayType or TupleType violates this rule.
 */
function checkArrayOrTupleType(node, context, options) {
    return {
        context,
        descriptors: (!node.parent ||
            !isTSTypeOperator(node.parent) ||
            node.parent.operator !== "readonly") &&
            (!options.ignoreReturnType || !isInReturnType(node))
            ? [
                {
                    node,
                    messageId: isTSTupleType(node) ? "tuple" : "array",
                    fix: node.parent && isTSArrayType(node.parent)
                        ? fixer => [
                            fixer.insertTextBefore(node, "(readonly "),
                            fixer.insertTextAfter(node, ")")
                        ]
                        : fixer => fixer.insertTextBefore(node, "readonly ")
                }
            ]
            : []
    };
}
/**
 * Check if the given TypeReference violates this rule.
 */
function checkTypeReference(node, context, options) {
    if (isIdentifier(node.typeName)) {
        const immutableType = mutableToImmutableTypes.get(node.typeName.name);
        return {
            context,
            descriptors: immutableType && (!options.ignoreReturnType || !isInReturnType(node))
                ? [
                    {
                        node,
                        messageId: "type",
                        fix: fixer => fixer.replaceText(node.typeName, immutableType)
                    }
                ]
                : []
        };
    }
    else {
        return {
            context,
            descriptors: []
        };
    }
}
/**
 * Check if the given property/signature node violates this rule.
 */
function checkProperty(node, context) {
    return {
        context,
        descriptors: node.readonly
            ? []
            : [
                {
                    node,
                    messageId: "property",
                    fix: isTSIndexSignature(node)
                        ? fixer => fixer.insertTextBefore(node, "readonly ")
                        : isTSParameterProperty(node)
                            ? fixer => fixer.insertTextBefore(node.parameter, "readonly ")
                            : fixer => fixer.insertTextBefore(node.key, "readonly ")
                }
            ]
    };
}
/**
 * Check if the given TypeReference violates this rule.
 */
function checkImplicitType(node, context) {
    const declarators = isFunctionLike(node)
        ? node.params
            .map(param => isAssignmentPattern(param)
            ? /* eslint-disable-next-line @typescript-eslint/no-object-literal-type-assertion */
                { id: param.left, init: param.right, node: param }
            : undefined)
            .filter((param) => param !== undefined)
        : node.declarations.map(declaration => 
        /* eslint-disable-next-line @typescript-eslint/no-object-literal-type-assertion */
        ({
            id: declaration.id,
            init: declaration.init,
            node: declaration
        }));
    return {
        context,
        descriptors: declarators.flatMap(declarator => {
            return isIdentifier(declarator.id) &&
                declarator.id.typeAnnotation === undefined &&
                declarator.init !== null &&
                isArrayType(getTypeOfNode(declarator.init, context))
                ? [
                    {
                        node: declarator.node,
                        messageId: "implicit",
                        fix: fixer => fixer.insertTextAfter(declarator.id, ": readonly unknown[]")
                    }
                ]
                : [];
        })
    };
}
// Create the rule.
const rule$e = createRule(name$e, meta$e, defaultOptions$e, (context, options) => {
    const _checkArrayOrTupleType = checkNode(checkArrayOrTupleType, context, options);
    const _checkTypeReference = checkNode(checkTypeReference, context, options);
    const _checkProperty = checkNode(checkProperty, context, options);
    const _checkImplicitType = checkNode(checkImplicitType, context, options);
    return Object.assign({ TSArrayType: _checkArrayOrTupleType, TSTupleType: _checkArrayOrTupleType, TSTypeReference: _checkTypeReference, ClassProperty: _checkProperty, TSIndexSignature: _checkProperty, TSParameterProperty: _checkProperty, TSPropertySignature: _checkProperty }, (options.checkImplicit
        ? {
            VariableDeclaration: _checkImplicitType,
            FunctionDeclaration: _checkImplicitType,
            FunctionExpression: _checkImplicitType,
            ArrowFunctionExpression: _checkImplicitType
        }
        : {}));
});

/**
 * All of the custom rules.
 */
const rules = {
    [name]: rule,
    [name$1]: rule$1,
    [name$2]: rule$2,
    [name$4]: rule$4,
    [name$5]: rule$5,
    [name$6]: rule$6,
    [name$7]: rule$7,
    [name$8]: rule$8,
    [name$9]: rule$9,
    [name$a]: rule$a,
    [name$3]: rule$3,
    [name$b]: rule$b,
    [name$c]: rule$c,
    [name$d]: rule$d,
    [name$e]: rule$e
};

const config$5 = {
    rules,
    configs: {
        all: config,
        recommended: config$2,
        functional: config$4,
        "functional-lite": config$3
    }
};

export default config$5;
